"""Owner related models"""
import datetime as dt
import json
import os
import re
from email.utils import format_datetime, parsedate_to_datetime
from typing import Optional

from bravado.exception import HTTPForbidden, HTTPNotFound

from django.contrib.auth.models import Group, User
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, transaction
from django.db.models import F
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.datetime import DATETIME_FORMAT
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..app_settings import (
    STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED,
    STRUCTURES_DEVELOPER_MODE,
    STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES,
    STRUCTURES_FEATURE_CUSTOMS_OFFICES,
    STRUCTURES_FEATURE_STARBASES,
    STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION,
    STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES,
    STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED,
    STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES,
)
from ..constants import EveGroupId, EveTypeId
from ..managers import OwnerManager
from ..providers import esi
from .eveuniverse import EveSovereigntyMap
from .notifications import (
    EveEntity,
    GeneratedNotification,
    Notification,
    NotificationType,
    Webhook,
)
from .structures import (
    PocoDetails,
    StarbaseDetail,
    StarbaseDetailFuel,
    Structure,
    StructureItem,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class General(models.Model):
    """Meta model for global app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and view public pages"),
            ("view_corporation_structures", "Can view corporation structures"),
            ("view_alliance_structures", "Can view alliance structures"),
            ("view_all_structures", "Can view all structures"),
            ("add_structure_owner", "Can add new structure owner"),
            (
                "view_all_unanchoring_status",
                "Can view unanchoring timers for all structures the user can see",
            ),
            ("view_structure_fit", "Can view structure fit"),
        )


class Owner(models.Model):
    """A corporation that owns structures"""

    class RotateCharactersType(models.TextChoices):
        """Type of sync to rotate characters for."""

        STRUCTURES = "structures"
        NOTIFICATIONS = "notifications"

        @property
        def last_used_at_name(self) -> str:
            """Name of last used at property."""
            if self is self.STRUCTURES:
                return "structures_last_used_at"
            elif self is self.NOTIFICATIONS:
                return "notifications_last_used_at"
            raise NotImplementedError(f"Not defined for: {self}")

        @property
        def esi_cache_duration(self) -> int:
            """ESI cache duration in seconds."""
            if self is self.STRUCTURES:
                return 3600
            elif self is self.NOTIFICATIONS:
                return 600
            raise NotImplementedError(f"Not defined for: {self}")

    # PK
    corporation = models.OneToOneField(
        EveCorporationInfo,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name="structure_owner",
        help_text="Corporation owning structures",
    )
    # regular
    are_pocos_public = models.BooleanField(
        default=False,
        help_text=("whether pocos of this owner are shown on public POCO page"),
    )
    assets_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text="when the last successful update happened",
    )
    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
        help_text="OUTDATED. Has been replaced by OwnerCharacter",
    )
    forwarding_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text="when the last successful update happened",
    )
    has_default_pings_enabled = models.BooleanField(
        default=True,
        help_text=(
            "to enable or disable pinging of notifications for this owner "
            "e.g. with @everyone and @here"
        ),
    )
    is_active = models.BooleanField(
        default=True,
        help_text=("whether this owner is currently included in the sync process"),
    )
    is_alliance_main = models.BooleanField(
        default=False,
        help_text=(
            "whether alliance wide notifications "
            "are forwarded for this owner (e.g. sov notifications)"
        ),
    )
    is_included_in_service_status = models.BooleanField(
        default=True,
        help_text=(
            "whether the sync status of this owner is included in "
            "the overall status of this services"
        ),
    )
    is_up = models.BooleanField(
        null=True,
        default=None,
        editable=False,
        help_text="whether all services for this owner are currently up",
    )
    notifications_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text="when the last successful update happened",
    )
    ping_groups = models.ManyToManyField(
        Group,
        default=None,
        blank=True,
        related_name="+",
        help_text="Groups to be pinged for each notification. ",
    )
    structures_last_update_at = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        help_text="when the last successful update happened",
    )
    webhooks = models.ManyToManyField(
        "Webhook",
        default=None,
        blank=True,
        related_name="owners",
        help_text="Notifications are sent to these webhooks. ",
    )

    objects = OwnerManager()

    def __str__(self) -> str:
        return str(self.corporation.corporation_name)

    def __repr__(self):
        return "{}(pk={}, corporation='{}')".format(
            self.__class__.__name__, self.pk, self.corporation
        )

    def save(self, *args, **kwargs) -> None:
        if self.is_alliance_main:
            Owner.objects.filter(
                corporation__alliance_id=self.corporation.alliance_id
            ).exclude(corporation__alliance_id__isnull=True).update(
                is_alliance_main=False
            )
            if "update_fields" in kwargs:
                kwargs["update_fields"].append("is_alliance_main")
        super().save(*args, **kwargs)

    @property
    def is_structure_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return self.structures_last_update_at and self.structures_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES)
        )

    @property
    def is_notification_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return (
            self.notifications_last_update_at
            and self.notifications_last_update_at
            > (now() - dt.timedelta(minutes=STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES))
        )

    @property
    def is_forwarding_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return self.forwarding_last_update_at and self.forwarding_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES)
        )

    @property
    def is_assets_sync_fresh(self) -> bool:
        """True if last sync happened with grace time, else False."""
        return self.assets_last_update_at and self.assets_last_update_at > (
            now() - dt.timedelta(minutes=STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES)
        )

    @property
    def are_all_syncs_ok(self) -> bool:
        """returns true if they have been no errors
        and last syncing occurred within alloted time for all sync categories
        """
        return (
            self.is_structure_sync_fresh
            and self.is_notification_sync_fresh
            and self.is_forwarding_sync_fresh
            and self.is_assets_sync_fresh
        )

    def update_is_up(self) -> bool:
        """Check if all services for this owner are up, notify admins if necessary
        and store result.
        """

        def fresh_str(value: bool) -> str:
            return "up" if value else "down"

        is_up = self.are_all_syncs_ok
        if (
            STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED
            and self.is_included_in_service_status
        ):
            if self.is_up and is_up is False:
                title = f"{__title__}: Services are down for {self}"
                msg = (
                    f"Structure services for {self} are down. "
                    "Admin action is likely required to restore services.\n"
                    f"- Structures: {fresh_str(self.is_structure_sync_fresh)}\n"
                    f"- Notifications: {fresh_str(self.is_notification_sync_fresh)}\n"
                    f"- Forwarding: {fresh_str(self.is_forwarding_sync_fresh)}\n"
                    f"- Assets: {fresh_str(self.is_assets_sync_fresh)}\n"
                )
                notify_admins(message=msg, title=title, level="danger")
            elif self.is_up is False and is_up:
                title = f"{__title__}: Services restored for {self}"
                msg = f"Structure services for {self} have been restored. "
                notify_admins(message=msg, title=title, level="success")
            elif self.is_up is None and is_up:
                title = f"{__title__}: Services enabled {self}"
                msg = f"Structure services for {self} have been enabled. "
                notify_admins(message=msg, title=title, level="success")
        if self.is_up != is_up:
            self.is_up = is_up
            self.save(update_fields=["is_up"])
        return is_up

    def add_character(
        self, character_ownership: CharacterOwnership
    ) -> "OwnerCharacter":
        """Add character to this owner.

        Raises ValueError when character does not belong to owner's corporation.
        """
        if (
            character_ownership.character.corporation_id
            != self.corporation.corporation_id
        ):
            raise ValueError(
                f"Character {character_ownership.character} does not belong "
                "to owner corporation."
            )
        obj, _ = self.characters.get_or_create(character_ownership=character_ownership)
        return obj

    def characters_count(self) -> int:
        """Count of valid owner characters."""
        return self.characters.count()

    def has_sov(self, eve_solar_system: EveSolarSystem) -> bool:
        """Determine whether this owner has sov in the given solar system."""
        return EveSovereigntyMap.objects.corporation_has_sov(
            eve_solar_system=eve_solar_system, corporation=self.corporation
        )

    def delete_character(
        self,
        character: "OwnerCharacter",
        error: str,
        level: str = "warning",
        max_allowed_errors: int = 0,
    ) -> None:
        """Delete character and notify it's owner and admin about the reason

        Args:
        - character: Character this error refers to
        - error: Error text
        - level: context level for the notification
        - max_error: how many errors are permitted before character is deleted
        """
        if character.error_count < max_allowed_errors:
            logger.warning(
                (
                    "%s: Character encountered an error and will be deleted "
                    "if this occurs more often (%d/%d): %s"
                ),
                character,
                character.error_count + 1,
                max_allowed_errors,
                error,
            )
            with transaction.atomic():
                character.error_count = F("error_count") + 1
                character.save(update_fields=["error_count"])
            return

        title = f"{__title__}: {self}: Invalid character has been removed"
        message = (
            f"{character.character_ownership}: {error}\n"
            "The character has been removed. "
            "Please add a new character to restore the previous service level."
        )
        notify(
            user=character.character_ownership.user,
            title=title,
            message=message,
            level=level,
        )
        if self.characters.count() == 1:
            message += (
                " This owner has no configured characters anymore "
                "and it's services are now down."
            )
            level = "danger"
        notify_admins(title=f"FYI: {title}", message=message, level=level)
        character.delete()

    def _rotate_character(
        self,
        character: "OwnerCharacter",
        ignore_schedule: bool,
        rotate_characters: RotateCharactersType,
    ) -> None:
        """Rotate this character such that all are spread evenly
        across the ESI cache duration for each ESI call.
        """
        time_since_last_used = (
            (
                now() - getattr(character, rotate_characters.last_used_at_name)
            ).total_seconds()
            if getattr(character, rotate_characters.last_used_at_name)
            else None
        )
        try:
            minimum_time_between_rotations = max(
                rotate_characters.esi_cache_duration / self.characters.count(),
                60,
            )
        except ZeroDivisionError:
            minimum_time_between_rotations = rotate_characters.esi_cache_duration
        if (
            ignore_schedule
            or not time_since_last_used
            or time_since_last_used >= minimum_time_between_rotations
        ):
            setattr(character, rotate_characters.last_used_at_name, now())
            character.save()

    def fetch_token(
        self,
        rotate_characters: RotateCharactersType = None,
        ignore_schedule: bool = False,
    ) -> Token:
        """Fetch a valid token for the owner and return it.

        Args:
            rotate_characters: For which sync type to rotate through characters \
                with every new call
            ignore_schedule: Ignore current schedule when rotating

        Raises TokenError when no valid token can be provided.
        """
        token = None
        order_by_last_used = (
            rotate_characters.last_used_at_name
            if rotate_characters
            else "notifications_last_used_at"
        )
        for character in self.characters.order_by(order_by_last_used):
            if (
                character.character_ownership.character.corporation_id
                != self.corporation.corporation_id
            ):
                self.delete_character(
                    character=character,
                    error="Character does no longer belong to the owner's corporation.",
                )
                continue
            elif not character.character_ownership.user.has_perm(
                "structures.add_structure_owner"
            ):
                self.delete_character(
                    character=character,
                    error="Character does not have sufficient permission to sync.",
                )
                continue
            token = character.valid_token()
            if not token:
                self.delete_character(
                    character=character,
                    error="Character has no valid token for sync.",
                )
                continue
            break  # leave the for loop if we have found a valid token

        if not token:
            error = (
                f"{self}: No valid character found for sync. "
                "Service down for this owner."
            )
            raise TokenError(error)
        if rotate_characters:
            self._rotate_character(
                character=character,
                ignore_schedule=ignore_schedule,
                rotate_characters=rotate_characters,
            )
        return token

    def _report_esi_issue(self, action: str, ex: Exception, token: Token):
        """Report an ESI issue to admins."""
        message = f"{self}: Failed to {action} from ESI with token {token} due to {ex}"
        logger.warning(message, exc_info=True)

    def update_structures_esi(self, user: User = None):
        """Updates all structures from ESI."""
        token = self.fetch_token(rotate_characters=self.RotateCharactersType.STRUCTURES)
        is_ok = self._fetch_upwell_structures(token)
        if STRUCTURES_FEATURE_CUSTOMS_OFFICES:
            is_ok &= self._fetch_custom_offices(token)
        if STRUCTURES_FEATURE_STARBASES:
            is_ok &= self._fetch_starbases(token)

        if is_ok:
            self.structures_last_update_at = now()
            self.save(update_fields=["structures_last_update_at"])
            if user:
                self._send_report_to_user(
                    topic="structures", topic_count=self.structures.count(), user=user
                )

    def _remove_structures_not_returned_from_esi(
        self, structures_qs: models.QuerySet, new_structures: list
    ):
        """Remove structures no longer returned from ESI."""
        ids_local = {x.id for x in structures_qs}
        ids_from_esi = {x["structure_id"] for x in new_structures}
        ids_to_remove = ids_local - ids_from_esi
        if len(ids_to_remove) > 0:
            structures_qs.filter(id__in=ids_to_remove).delete()
            logger.info(
                "Removed %d structures which apparently no longer exist.",
                len(ids_to_remove),
            )

    def _fetch_locations_for_assets(self, item_ids: list, token: Token) -> dict:
        """Fetch locations for given asset items from ESI."""
        logger.info(
            "%s: Fetching locations for %d assets from ESI", self, len(item_ids)
        )
        locations_data = list()
        for item_ids_chunk in chunks(list(item_ids), 999):
            try:
                locations_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_locations(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                locations_data += locations_data_chunk
        positions = {x["item_id"]: x["position"] for x in locations_data}
        return positions

    def _fetch_upwell_structures(self, token: Token) -> bool:
        """Fetch Upwell structures from ESI for self.

        Return True if successful, else False.
        """
        is_ok = True
        # fetch main list of structure for this corporation
        try:
            structures = (
                esi.client.Corporation.get_corporations_corporation_id_structures(
                    corporation_id=self.corporation.corporation_id,
                    token=token.valid_access_token(),
                ).results()
            )
        except OSError as ex:
            self._report_esi_issue("fetch corporation structures", ex, token)
            return False

        # fetch additional information for structures
        if not structures:
            logger.info("%s: No Upwell structures retrieved from ESI", self)
        else:
            logger.info(
                "%s: Fetching additional infos for %d Upwell structures from ESI",
                self,
                len(structures),
            )
            for structure in structures:
                try:
                    structure_info = (
                        esi.client.Universe.get_universe_structures_structure_id(
                            structure_id=structure["structure_id"],
                            token=token.valid_access_token(),
                        )
                    ).results()
                except OSError as ex:
                    self._report_esi_issue(
                        f"fetch structure #{structure['structure_id']}", ex, token
                    )
                    structure["name"] = "(no data)"
                    is_ok = False
                else:
                    structure["name"] = Structure.extract_name_from_esi_response(
                        structure_info["name"]
                    )
                    structure["position"] = structure_info["position"]

            logger.info(
                "%s: Storing updates for %d upwell structures",
                self,
                len(structures),
            )
            for structure in structures:
                try:
                    Structure.objects.update_or_create_from_dict(structure, self)
                except OSError:
                    logger.warning(
                        "%s: Failed to store update for structure with ID %s",
                        self,
                        structure["structure_id"],
                    )
                    is_ok = False

        if STRUCTURES_DEVELOPER_MODE:
            self._store_raw_data("structures", structures)

        self._remove_structures_not_returned_from_esi(
            structures_qs=self.structures.filter_upwell_structures(),
            new_structures=structures,
        )
        return is_ok

    def _fetch_custom_offices(self, token: Token) -> bool:
        """Fetch custom offices from ESI for this owner.

        Return True when successful, else False.
        """
        structures = dict()
        try:
            pocos = esi.client.Planetary_Interaction.get_corporations_corporation_id_customs_offices(
                corporation_id=self.corporation.corporation_id,
                token=token.valid_access_token(),
            ).results()
            if not pocos:
                logger.info("%s: No custom offices retrieved from ESI", self)
            else:
                pocos_2 = {row["office_id"]: row for row in pocos}
                office_ids = list(pocos_2.keys())
                positions = self._fetch_locations_for_assets(office_ids, token)
                names = self._fetch_names_for_pocos(office_ids, token)

                # making sure we have all solar systems loaded
                # incl. their planets for later name matching
                for solar_system_id in {int(x["system_id"]) for x in pocos}:
                    EveSolarSystem.objects.get_or_create_esi(id=solar_system_id)

                # compile pocos into structures list
                for office_id, poco in pocos_2.items():
                    planet_name = names.get(office_id, "")
                    if planet_name:
                        try:
                            eve_planet = EvePlanet.objects.get(name=planet_name)
                        except EvePlanet.DoesNotExist:
                            name = ""
                            planet_id = None
                        else:
                            planet_id = eve_planet.id
                            name = eve_planet.eve_type.name
                    else:
                        name = None
                        planet_id = None

                    reinforce_exit_start = dt.datetime(
                        year=2000, month=1, day=1, hour=poco["reinforce_exit_start"]
                    )
                    reinforce_hour = reinforce_exit_start + dt.timedelta(hours=1)
                    structure = {
                        "structure_id": office_id,
                        "type_id": EveTypeId.CUSTOMS_OFFICE,
                        "corporation_id": self.corporation.corporation_id,
                        "name": name if name else "",
                        "system_id": poco["system_id"],
                        "reinforce_hour": reinforce_hour.hour,
                        "state": Structure.State.UNKNOWN,
                    }
                    if planet_id:
                        structure["planet_id"] = planet_id

                    if office_id in positions:
                        structure["position"] = positions[office_id]

                    structures[office_id] = structure

                logger.info(
                    "%s: Storing updates for %d customs offices", self, len(structures)
                )
                for office_id, structure in structures.items():
                    structure_obj, _ = Structure.objects.update_or_create_from_dict(
                        structure, self
                    )
                    try:
                        poco = pocos_2[office_id]
                    except KeyError:
                        logger.warning(
                            "%s: No details found for this POCO: %d", self, office_id
                        )
                    else:
                        standing_level = PocoDetails.StandingLevel.from_esi(
                            poco.get("standing_level")
                        )
                        PocoDetails.objects.update_or_create(
                            structure=structure_obj,
                            defaults={
                                "alliance_tax_rate": poco.get("alliance_tax_rate"),
                                "allow_access_with_standings": poco.get(
                                    "allow_access_with_standings"
                                ),
                                "allow_alliance_access": poco.get(
                                    "allow_alliance_access"
                                ),
                                "bad_standing_tax_rate": poco.get(
                                    "bad_standing_tax_rate"
                                ),
                                "corporation_tax_rate": poco.get(
                                    "corporation_tax_rate"
                                ),
                                "excellent_standing_tax_rate": poco.get(
                                    "excellent_standing_tax_rate"
                                ),
                                "good_standing_tax_rate": poco.get(
                                    "good_standing_tax_rate"
                                ),
                                "neutral_standing_tax_rate": poco.get(
                                    "neutral_standing_tax_rate"
                                ),
                                "reinforce_exit_end": poco.get("reinforce_exit_end"),
                                "reinforce_exit_start": poco.get(
                                    "reinforce_exit_start"
                                ),
                                "standing_level": standing_level,
                                "terrible_standing_tax_rate": poco.get(
                                    "terrible_standing_tax_rate"
                                ),
                            },
                        )

            if STRUCTURES_DEVELOPER_MODE:
                self._store_raw_data("customs_offices", structures)

        except OSError as ex:
            self._report_esi_issue("fetch custom offices", ex, token)
            return False

        self._remove_structures_not_returned_from_esi(
            structures_qs=self.structures.filter_customs_offices(),
            new_structures=structures.values(),
        )
        return True

    def _fetch_names_for_pocos(self, item_ids: list, token: Token) -> dict:
        logger.info(
            "%s: Fetching names for %d custom office names from ESI",
            self,
            len(item_ids),
        )
        names_data = list()
        for item_ids_chunk in chunks(item_ids, 999):
            try:
                names_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_names(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                names_data += names_data_chunk
        names = {x["item_id"]: self._extract_planet_name(x["name"]) for x in names_data}
        return names

    @staticmethod
    def _extract_planet_name(text: str) -> str:
        """Extract name of planet from assert name for a customs office."""
        reg_ex = re.compile(r"Customs Office \((.+)\)")
        matches = reg_ex.match(text)
        return matches.group(1) if matches else ""

    def _fetch_starbases(self, token: Token) -> bool:
        """Fetch starbases from ESI for this owner.

        Return True when successful, else False.
        """
        structures = list()
        try:
            starbases_data = (
                esi.client.Corporation.get_corporations_corporation_id_starbases(
                    corporation_id=self.corporation.corporation_id,
                    token=token.valid_access_token(),
                )
            ).results()
            if not starbases_data:
                logger.info("%s: This corporation has no starbases.", self)
            else:
                starbases_data = {obj["starbase_id"]: obj for obj in starbases_data}
                names = self._fetch_starbases_names(starbases_data.keys(), token)
                locations = self._fetch_locations_for_assets(
                    starbases_data.keys(), token
                )
                # convert starbases to structures
                for structure_id, starbase in starbases_data.items():
                    try:
                        name = names[structure_id]
                    except KeyError:
                        name = "Starbase"
                    structure = {
                        "structure_id": structure_id,
                        "type_id": starbase["type_id"],
                        "corporation_id": self.corporation.corporation_id,
                        "name": name,
                        "system_id": starbase["system_id"],
                    }
                    if structure_id in locations:
                        structure["position"] = locations[structure_id]
                    if "state" in starbase:
                        structure["state"] = starbase["state"]
                    if "moon_id" in starbase:
                        structure["moon_id"] = starbase["moon_id"]
                    if "fuel_expires" in starbase:
                        structure["fuel_expires"] = starbase["fuel_expires"]
                    if "reinforced_until" in starbase:
                        structure["state_timer_end"] = starbase["reinforced_until"]
                    if "unanchors_at" in starbase:
                        structure["unanchors_at"] = starbase["unanchors_at"]
                    structures.append(structure)

                logger.info(
                    "%s: Storing updates for %d starbases", self, len(structures)
                )
                for structure in structures:
                    structure_obj, _ = Structure.objects.update_or_create_from_dict(
                        structure, self
                    )
                    detail = self._update_starbase_detail(
                        structure=structure_obj, token=token
                    )
                    fuel_expires_at = detail.calc_fuel_expires()
                    if fuel_expires_at:
                        structure_obj.fuel_expires_at = fuel_expires_at
                        structure_obj.save()
                    if (
                        structure_obj.state == Structure.State.POS_REINFORCED
                        and structure_obj.state_timer_end
                    ):
                        GeneratedNotification.objects.get_or_create_from_structure(
                            structure=structure_obj,
                            notif_type=NotificationType.TOWER_REINFORCED_EXTRA,
                        )

            if STRUCTURES_DEVELOPER_MODE:
                self._store_raw_data("starbases", structures)

        except HTTPForbidden:
            try:
                character = self.characters.get(
                    character_ownership__character__character_id=token.character_id
                )
            except ObjectDoesNotExist:
                pass
            else:
                self.delete_character(
                    character=character,
                    error=(
                        "Character is not a director or CEO and therefore "
                        "can not fetch starbases."
                    ),
                    max_allowed_errors=STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES,
                )
            return False

        except OSError as ex:
            self._report_esi_issue("fetch starbases", ex, token)
            return False

        self._remove_structures_not_returned_from_esi(
            structures_qs=self.structures.filter_starbases(),
            new_structures=structures,
        )
        return True

    def _fetch_starbases_names(self, item_ids: list, token: Token) -> dict:
        logger.info("%s: Fetching names for %d starbases from ESI", self, len(item_ids))
        names_data = list()
        for item_ids_chunk in chunks(list(item_ids), 999):
            try:
                names_data_chunk = (
                    esi.client.Assets.post_corporations_corporation_id_assets_names(
                        corporation_id=self.corporation.corporation_id,
                        item_ids=item_ids_chunk,
                        token=token.valid_access_token(),
                    )
                ).results()
            except HTTPNotFound:
                pass
            else:
                names_data += names_data_chunk
        names = {x["item_id"]: x["name"] for x in names_data}
        return names

    def _update_starbase_detail(
        self, structure: object, token: Token
    ) -> StarbaseDetail:
        """Update detail for the starbase from ESI."""
        operation = esi.client.Corporation.get_corporations_corporation_id_starbases_starbase_id(
            corporation_id=structure.owner.corporation.corporation_id,
            starbase_id=structure.id,
            system_id=structure.eve_solar_system.id,
            token=token.valid_access_token(),
        )
        operation.request_config.also_return_response = True
        data, response = operation.results()
        last_modified_at = parsedate_to_datetime(
            response.headers.get("Last-Modified", format_datetime(now()))
        )
        defaults = {
            "allow_alliance_members": data["allow_alliance_members"],
            "allow_corporation_members": data["allow_corporation_members"],
            "anchor_role": StarbaseDetail.Role.from_esi(data["anchor"]),
            "attack_if_at_war": data["attack_if_at_war"],
            "attack_if_other_security_status_dropping": data.get(
                "attack_if_other_security_status_dropping"
            ),
            "attack_security_status_threshold": data.get(
                "attack_security_status_threshold"
            ),
            "attack_standing_threshold": data.get("attack_standing_threshold"),
            "fuel_bay_take_role": StarbaseDetail.Role.from_esi(data["fuel_bay_take"]),
            "fuel_bay_view_role": StarbaseDetail.Role.from_esi(data["fuel_bay_view"]),
            "last_modified_at": last_modified_at,
            "offline_role": StarbaseDetail.Role.from_esi(data["offline"]),
            "online_role": StarbaseDetail.Role.from_esi(data["online"]),
            "unanchor_role": StarbaseDetail.Role.from_esi(data["unanchor"]),
            "use_alliance_standings": data["use_alliance_standings"],
        }
        for fuel in data["fuels"]:
            EveType.objects.get_or_create_esi(id=fuel["type_id"])
        with transaction.atomic():
            detail, _ = StarbaseDetail.objects.update_or_create(
                structure=structure, defaults=defaults
            )
            detail.fuels.all().delete()
            for fuel in data["fuels"]:
                StarbaseDetailFuel.objects.create(
                    eve_type_id=fuel["type_id"],
                    detail=detail,
                    quantity=fuel["quantity"],
                )
        return detail

    def add_or_remove_timers_from_notifications(self):
        """Add/remove timers from esi and generated notification of this owner."""
        self.notification_set.add_or_remove_timers()
        self.generatednotification_set.add_or_remove_timers()

    def fetch_notifications_esi(self, user: User = None) -> None:
        """Fetch notifications for this owner from ESI and process them."""
        notifications_count_all = 0
        token = self.fetch_token(
            rotate_characters=self.RotateCharactersType.NOTIFICATIONS
        )
        notifications = self._fetch_notifications_from_esi(token)
        notifications_count_new = self._store_notifications(notifications)
        self._process_moon_notifications()
        if notifications_count_new > 0:
            logger.info(
                "%s: Received %d new notifications from ESI",
                self,
                notifications_count_new,
            )
            notifications_count_all += notifications_count_new
        else:
            logger.info("%s: No new notifications received from ESI", self)

        self.notifications_last_update_at = now()
        self.save(update_fields=["notifications_last_update_at"])
        if user:
            self._send_report_to_user(
                topic="notifications",
                topic_count=notifications_count_all,
                user=user,
            )

    def _fetch_notifications_from_esi(self, token: Token) -> dict:
        """fetching all notifications from ESI for current owner"""
        notifications = esi.client.Character.get_characters_character_id_notifications(
            character_id=token.character_id, token=token.valid_access_token()
        ).results()
        if STRUCTURES_DEVELOPER_MODE:
            self._store_raw_data("notifications", notifications)
        if STRUCTURES_NOTIFICATIONS_ARCHIVING_ENABLED:
            self._store_raw_notifications(notifications)
        logger.debug(
            "%s: Processing %d notifications received from ESI",
            self,
            len(notifications),
        )
        return notifications

    def _store_raw_notifications(self, notifications):
        # store notifications to disk in continuous file per corp
        folder_name = "structures_notifications_archive"
        os.makedirs(folder_name, exist_ok=True)
        filename = "{}/notifications_{}_{}.txt".format(
            folder_name, self.corporation.corporation_id, now().date().isoformat()
        )
        logger.info(
            "%s: Storing notifications into archive file: %s", self, format(filename)
        )
        with open(file=filename, mode="a", encoding="utf-8") as f:
            f.write(
                "[{}] {}:\n".format(
                    now().strftime(DATETIME_FORMAT),
                    self.corporation.corporation_ticker,
                )
            )
            json.dump(notifications, f, cls=DjangoJSONEncoder, sort_keys=True, indent=4)
            f.write("\n")

    def _store_notifications(self, notifications: list) -> int:
        """Store new notifications in database.
        Returns number of newly created objects.
        """
        # identify new notifications
        existing_notification_ids = set(
            self.notification_set.values_list("notification_id", flat=True)
        )
        new_notifications = [
            obj
            for obj in notifications
            if obj["notification_id"] not in existing_notification_ids
        ]
        # create new notification objects
        for notification in new_notifications:
            if notification["sender_type"] == "other":
                sender = None
            else:
                sender, _ = EveEntity.objects.get_or_create_esi(
                    id=notification["sender_id"]
                )
            text = notification["text"] if "text" in notification else None
            is_read = notification["is_read"] if "is_read" in notification else None
            # at least one type has a trailing white space
            # which we need to remove
            notif_type = notification["type"].strip()
            Notification.objects.create(
                notification_id=notification["notification_id"],
                owner=self,
                sender=sender,
                timestamp=notification["timestamp"],
                notif_type=notif_type,
                text=text,
                is_read=is_read,
                last_updated=now(),
                created=now(),
            )
        return len(new_notifications)

    def _process_moon_notifications(self):
        """processes notifications for timers if any"""
        empty_refineries = Structure.objects.filter(
            owner=self,
            eve_type__eve_group_id=EveGroupId.REFINERY,
            eve_moon__isnull=True,
        )
        if empty_refineries.exists():
            logger.info(
                "%s: Trying to find moons for up to %d refineries which have no moon.",
                self,
                empty_refineries.count(),
            )
            notifications = (
                self.notification_set.filter(
                    notif_type__in=NotificationType.relevant_for_moonmining
                )
                .select_related("owner", "sender")
                .order_by("timestamp")
            )
            structure_id_2_moon_id = dict()
            for notification in notifications:
                parsed_text = notification.parsed_text()
                moon_id = parsed_text["moonID"]
                structure_id = parsed_text["structureID"]
                structure_id_2_moon_id[structure_id] = moon_id
            for refinery in empty_refineries:
                if refinery.id in structure_id_2_moon_id:
                    logger.info("%s: Updating moon for structure %s", self, refinery)
                    eve_moon, _ = EveMoon.objects.get_or_create_esi(
                        id=structure_id_2_moon_id[refinery.id]
                    )
                    refinery.eve_moon = eve_moon
                    refinery.save()

    def send_new_notifications(self, user: User = None):
        """Forward all new notification of this owner to configured webhooks."""
        notifications_count = 0
        cutoff_dt_for_stale = now() - dt.timedelta(
            hours=STRUCTURES_HOURS_UNTIL_STALE_NOTIFICATION
        )
        my_filter = {
            "notif_type__in": (
                Webhook.objects.enabled_notification_types()
                & NotificationType.relevant_for_forwarding
            ),
            "is_sent": False,
            "timestamp__gte": cutoff_dt_for_stale,
        }
        new_eve_notifications = (
            self.notification_set.filter(**my_filter)
            .select_related("owner", "sender", "owner__corporation")
            .order_by("timestamp")
        )
        for notif in new_eve_notifications:
            notif.send_to_configured_webhooks()
        new_generated_notifications = (
            self.generatednotification_set.filter(**my_filter)
            .select_related("owner", "owner__corporation")
            .order_by("timestamp")
        )
        for notif in new_generated_notifications:
            notif.send_to_configured_webhooks()
        if (
            not new_eve_notifications.exists()
            and not new_generated_notifications.exists()
        ):
            logger.info("%s: No new notifications found for forwarding", self)
        self.forwarding_last_update_at = now()
        self.save(update_fields=["forwarding_last_update_at"])
        if user:
            self._send_report_to_user(
                topic="notifications", topic_count=notifications_count, user=user
            )

    def _send_report_to_user(self, topic: str, topic_count: int, user: User):
        message_details = "%(count)s %(topic)s synced." % {
            "count": topic_count,
            "topic": topic,
        }
        message = _(
            "Syncing of %(topic)s for %(owner)s %(result)s.\n %(message_details)s"
        ) % {
            "topic": topic,
            "owner": self.corporation.corporation_name,
            "result": _("completed successfully"),
            "message_details": message_details,
        }
        notify(
            user,
            title=_("%(title)s: %(topic)s updated for %(owner)s: %(result)s")
            % {
                "title": _(__title__),
                "topic": topic,
                "owner": self.corporation.corporation_name,
                "result": _("OK"),
            },
            message=message,
            level="success",
        )

    def _store_raw_data(self, name: str, data: list):
        """store raw data for debug purposes"""
        with open(
            f"{name}_raw_{self.corporation.corporation_id}.json", "w", encoding="utf-8"
        ) as f:
            json.dump(data, f, cls=DjangoJSONEncoder, sort_keys=True, indent=4)

    def update_asset_esi(self, user: User = None):
        token = self.fetch_token()
        assets_data = self._fetch_structure_assets_from_esi(token)
        self._store_items_for_upwell_structures(assets_data)
        self._store_items_for_starbases(assets_data)
        if user:
            self._send_report_to_user(
                topic="assets", topic_count=self.structures.count(), user=user
            )

    def _fetch_structure_assets_from_esi(self, token: Token) -> dict:
        assets_raw = esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=self.corporation.corporation_id,
            token=token.valid_access_token(),
        ).results()
        assets = {asset["item_id"]: asset for asset in assets_raw}
        positions = self._fetch_locations_for_assets(assets.keys(), token)
        for item_id, asset in assets.items():
            asset["position"] = positions[item_id] if item_id in positions else None
        return assets

    def _store_items_for_upwell_structures(self, assets_data: dict):
        structure_ids = set(
            self.structures.filter_upwell_structures().values_list("id", flat=True)
        )
        assets_in_structures = {id: dict() for id in structure_ids}
        for item_id, item in assets_data.items():
            location_id = item["location_id"]
            if location_id in structure_ids and item["location_flag"] not in [
                StructureItem.LocationFlag.CORP_DELIVERIES,
                StructureItem.LocationFlag.OFFICE_FOLDER,
                StructureItem.LocationFlag.SECONDARY_STORAGE,
                StructureItem.LocationFlag.AUTOFIT,
            ]:
                assets_in_structures[location_id][item_id] = item
        for structure in self.structures.all():
            structure_items = list()
            if structure.id in assets_in_structures.keys():
                structure_assets = assets_in_structures[structure.id]
                has_fitting = [
                    asset
                    for asset in structure_assets.values()
                    if asset["location_flag"]
                    != StructureItem.LocationFlag.QUANTUM_CORE_ROOM
                ]
                has_core = [
                    asset
                    for asset in structure_assets.values()
                    if asset["location_flag"]
                    == StructureItem.LocationFlag.QUANTUM_CORE_ROOM
                ]
                structure.has_fitting = bool(has_fitting)
                structure.has_core = bool(has_core)
                structure.save()

                for asset in structure_assets.values():
                    structure_items.append(
                        StructureItem.from_esi_asset(asset, structure)
                    )

            structure.update_items(structure_items)
            structure.reevaluate_jump_fuel_alerts()

        self.assets_last_update_at = now()
        self.save(update_fields=["assets_last_update_at"])

    def _store_items_for_starbases(self, assets_raw: dict):
        for structure in self.structures.filter_starbases().filter(
            position_x__isnull=False, position_y__isnull=False, position_z__isnull=False
        ):
            structure_items = list()
            for item_id, item in assets_raw.items():
                if (
                    item["location_id"] == structure.eve_solar_system_id
                    and item["location_type"] == "solar_system"
                    and item["location_flag"] == "AutoFit"
                    and item_id != structure.id
                    and item["position"]
                    and structure.distance_to_object(
                        item["position"]["x"],
                        item["position"]["y"],
                        item["position"]["z"],
                    )
                    < 100_000
                ):
                    structure_items.append(
                        StructureItem.from_esi_asset(item, structure)
                    )
            structure.update_items(structure_items)

    @classmethod
    def get_esi_scopes(cls) -> list:
        scopes = [
            "esi-corporations.read_structures.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_notifications.v1",
            "esi-assets.read_corporation_assets.v1",
        ]
        if STRUCTURES_FEATURE_CUSTOMS_OFFICES:
            scopes += ["esi-planets.read_customs_offices.v1"]
        if STRUCTURES_FEATURE_STARBASES:
            scopes += ["esi-corporations.read_starbases.v1"]
        return scopes


class OwnerCharacter(models.Model):
    """Character for syncing owner data with ESI."""

    owner = models.ForeignKey(
        Owner, on_delete=models.CASCADE, related_name="characters"
    )
    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.CASCADE,
        related_name="+",
        help_text="character used for syncing",
    )
    structures_last_used_at = models.DateTimeField(
        null=True,
        default=None,
        editable=False,
        db_index=True,
        help_text="when this character was last used for syncing structures",
    )
    notifications_last_used_at = models.DateTimeField(
        null=True,
        default=None,
        editable=False,
        db_index=True,
        help_text="when this character was last used for syncing notifications",
    )
    error_count = models.PositiveIntegerField(
        default=0,
        editable=False,
        help_text="Count of ESI errors which happened with this character.",
    )
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["owner", "character_ownership"], name="functional_pk_ownertoken"
            )
        ]

    def __str__(self) -> str:
        return (
            f"{self.owner.corporation.corporation_name}-"
            f"{self.character_ownership.character.character_name}"
        )

    def valid_token(self) -> Optional[Token]:
        """Provide a valid token or None if none can be found."""
        return (
            Token.objects.filter(
                user=self.character_ownership.user,
                character_id=self.character_ownership.character.character_id,
            )
            .require_scopes(Owner.get_esi_scopes())
            .require_valid()
            .first()
        )



# Generated by Django 2.2.8 on 2020-01-20 01:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0005_notification_created"),
    ]

    operations = [
        migrations.CreateModel(
            name="StructureTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="name of the tag - must be unique",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="description for this tag",
                        null=True,
                    ),
                ),
                (
                    "style",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("default", "grey"),
                            ("primary", "dark blue"),
                            ("success", "green"),
                            ("info", "light blue"),
                            ("warning", "yellow"),
                            ("danger", "red"),
                        ],
                        default="default",
                        help_text="color style of tag",
                        max_length=16,
                    ),
                ),
                (
                    "is_default",
                    models.BooleanField(
                        default=False,
                        help_text="if true this tag will automatically added to new structures",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="structure",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="list of tags for this structure",
                to="structures.StructureTag",
            ),
        ),
    ]



"""Eve Universe models"""

from enum import Enum

from django.db import models
from eveuniverse.models import EveSolarSystem

from ..managers import EveSovereigntyMapManager


class EveSovereigntyMap(models.Model):
    """Shows which alliance / corporation / faction owns a system

    Note: This model does not hold FKs to respective objects like
    EveSolarSystem to avoid having load all those object from ESI
    """

    solar_system_id = models.PositiveIntegerField(primary_key=True)
    alliance_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="alliance who holds sov for this system",
    )
    corporation_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="corporation who holds sov for this system",
    )
    faction_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        db_index=True,
        help_text="faction who holds sov for this system",
    )
    last_updated = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        help_text="When this object was last updated from ESI",
        db_index=True,
    )

    objects = EveSovereigntyMapManager()

    def __str__(self):
        return str(self.solar_system_id)

    def __repr__(self):
        return "{}(solar_system_id='{}')".format(
            self.__class__.__name__, self.solar_system_id
        )


class EveSpaceType(str, Enum):
    UNKNOWN = "unknown"
    HIGHSEC = "highsec"
    LOWSEC = "lowsec"
    NULLSEC = "nullsec"
    W_SPACE = "w-space"

    @classmethod
    def from_solar_system(cls, eve_solar_system: EveSolarSystem) -> "EveSpaceType":
        """returns the space type"""
        if eve_solar_system.is_null_sec:
            return cls.NULLSEC
        elif eve_solar_system.is_low_sec:
            return cls.LOWSEC
        elif eve_solar_system.is_high_sec:
            return cls.HIGHSEC
        elif eve_solar_system.is_w_space:
            return cls.W_SPACE
        return cls.UNKNOWN



# Generated by Django 4.0.7 on 2022-08-18 12:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0042_migrate_sender_category_other"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="structureservice",
            name="name_de",
        ),
        migrations.RemoveField(
            model_name="structureservice",
            name="name_ko",
        ),
        migrations.RemoveField(
            model_name="structureservice",
            name="name_ru",
        ),
        migrations.RemoveField(
            model_name="structureservice",
            name="name_zh",
        ),
    ]



# Generated by Django 3.1.10 on 2021-05-22 13:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0024_fix_name_fields_too_small"),
    ]

    operations = [
        migrations.AddField(
            model_name="owner",
            name="are_pocos_public",
            field=models.BooleanField(
                default=False,
                help_text="whether pocos of this owner are shown on public POCO page",
            ),
        ),
        migrations.CreateModel(
            name="PocoDetails",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("alliance_tax_rate", models.FloatField(default=None, null=True)),
                ("allow_access_with_standings", models.BooleanField()),
                ("allow_alliance_access", models.BooleanField()),
                ("bad_standing_tax_rate", models.FloatField(default=None, null=True)),
                ("corporation_tax_rate", models.FloatField(default=None, null=True)),
                (
                    "excellent_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                ("good_standing_tax_rate", models.FloatField(default=None, null=True)),
                (
                    "neutral_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                ("reinforce_exit_end", models.PositiveIntegerField()),
                ("reinforce_exit_start", models.PositiveIntegerField()),
                (
                    "standing_level",
                    models.IntegerField(
                        choices=[
                            (-99, "none"),
                            (-10, "terrible"),
                            (-5, "bad"),
                            (0, "neutral"),
                            (5, "good"),
                            (10, "excellent"),
                        ],
                        default=-99,
                    ),
                ),
                (
                    "terrible_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "structure",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="poco_details",
                        to="structures.structure",
                    ),
                ),
            ],
        ),
    ]



"""Creating or removing timers from notifications."""

import datetime as dt

from django.utils.translation import gettext
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.datetime import (
    DATETIME_FORMAT,
    ldap_time_2_datetime,
    ldap_timedelta_2_timedelta,
)
from app_utils.django import app_labels
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..app_settings import (
    STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED,
    STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
)
from ..constants import EveTypeId
from ..models import Notification, NotificationType, Structure
from . import sovereignty, starbases

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer

    has_auth_timers = True
else:
    has_auth_timers = False

if "structuretimers" in app_labels():
    from structuretimers.models import Timer

    has_structure_timers = True
else:
    has_structure_timers = False

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def add_or_remove_timer(notif: Notification) -> bool:
    """Add/remove a timer related to this notification for some types.

    Returns True when timers where added or removed, else False
    """
    parsed_text = notif.parsed_text()
    if notif.notif_type in [
        NotificationType.STRUCTURE_LOST_ARMOR,
        NotificationType.STRUCTURE_LOST_SHIELD,
    ]:
        timer_processed = _gen_timer_structure_reinforcement(notif, parsed_text)
    elif notif.notif_type == NotificationType.SOV_STRUCTURE_REINFORCED:
        timer_processed = _gen_timer_sov_reinforcements(notif, parsed_text)
    elif notif.notif_type == NotificationType.ORBITAL_REINFORCED:
        timer_processed = _gen_timer_orbital_reinforcements(notif, parsed_text)
    elif notif.notif_type in [
        NotificationType.MOONMINING_EXTRACTION_STARTED,
        NotificationType.MOONMINING_EXTRACTION_CANCELLED,
    ]:
        if not STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED:
            timer_processed = None
        else:
            timer_processed = _gen_timer_moon_extraction(notif, parsed_text)
    elif notif.notif_type == NotificationType.TOWER_REINFORCED_EXTRA:
        timer_processed = _gen_timer_tower_reinforcements(notif, parsed_text)
    else:
        raise NotImplementedError(
            f"Unsupported notification type for timers: {notif.notif_type}"
        )
    if timer_processed:
        logger.info("%s: Created timer for notification", notif.notification_id)
        notif.is_timer_added = True
        notif.save()
    return timer_processed


def _gen_timer_structure_reinforcement(notif: Notification, parsed_text: str) -> bool:
    """Generate timer for structure reinforcements"""
    token = notif.owner.fetch_token()
    structure_obj, _ = Structure.objects.get_or_create_esi(
        id=parsed_text["structureID"], token=token
    )
    eve_time = notif.timestamp + ldap_timedelta_2_timedelta(parsed_text["timeLeft"])
    timer_processed = False
    if has_auth_timers:
        details_map = {
            NotificationType.STRUCTURE_LOST_SHIELD: gettext("Armor timer"),
            NotificationType.STRUCTURE_LOST_ARMOR: gettext("Final timer"),
        }
        AuthTimer.objects.create(
            details=details_map.get(notif.notif_type, ""),
            system=structure_obj.eve_solar_system.name,
            planet_moon="",
            structure=structure_obj.eve_type.name,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if has_structure_timers:
        timer_map = {
            NotificationType.STRUCTURE_LOST_SHIELD: Timer.Type.ARMOR,
            NotificationType.STRUCTURE_LOST_ARMOR: Timer.Type.HULL,
        }
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=structure_obj.eve_solar_system,
            structure_type=structure_obj.eve_type,
            timer_type=timer_map.get(notif.notif_type),
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name=structure_obj.name,
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_sov_reinforcements(notif: Notification, parsed_text: str) -> bool:
    """Generate timer for sov reinforcements."""
    if not notif.owner.is_alliance_main:
        return False

    solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
        id=parsed_text["solarSystemID"]
    )
    structure_type_name = sovereignty.structure_type_name_from_event_type(
        parsed_text["campaignEventType"]
    )
    eve_time = ldap_time_2_datetime(parsed_text["decloakTime"])
    timer_processed = False
    if has_auth_timers:
        AuthTimer.objects.create(
            details=gettext("Sov timer"),
            system=solar_system.name,
            planet_moon="",
            structure=structure_type_name,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if has_structure_timers:
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        structure_type, _ = EveType.objects.get_or_create_esi(
            id=sovereignty.type_id_from_event_type(parsed_text["campaignEventType"])
        )
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=eve_solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            owner_name=notif.sender.name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_orbital_reinforcements(notif: Notification, parsed_text: str) -> bool:
    """Generate timer for orbital reinforcements."""
    solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
        id=parsed_text["solarSystemID"]
    )
    planet, _ = EvePlanet.objects.get_or_create_esi(id=parsed_text["planetID"])
    eve_time = ldap_time_2_datetime(parsed_text["reinforceExitTime"])
    timer_processed = False
    if has_auth_timers:
        AuthTimer.objects.create(
            details=gettext("Final timer"),
            system=solar_system.name,
            planet_moon=planet.name,
            structure="POCO",
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if has_structure_timers:
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        structure_type, _ = EveType.objects.get_or_create_esi(
            id=EveTypeId.CUSTOMS_OFFICE
        )
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=eve_solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            location_details=planet.name,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name="Customs Office",
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_moon_extraction(notif: Notification, parsed_text: str) -> bool:
    """Generate timer for moon mining extractions."""
    solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
        id=parsed_text["solarSystemID"]
    )
    moon, _ = EveMoon.objects.get_or_create_esi(id=parsed_text["moonID"])
    if "readyTime" in parsed_text:
        eve_time = ldap_time_2_datetime(parsed_text["readyTime"])
    else:
        eve_time = None
    details = gettext("Extraction ready")
    system = solar_system.name
    planet_moon = moon.name
    structure_type_name = "Moon Mining Cycle"
    objective = "Friendly"
    timer_processed = False

    if has_structure_timers:
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=parsed_text["solarSystemID"]
        )
        structure_type, _ = EveType.objects.get_or_create_esi(
            id=parsed_text["structureTypeID"]
        )
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
    else:
        eve_solar_system = None
        structure_type = None
        visibility = None

    if notif.notif_type == NotificationType.MOONMINING_EXTRACTION_STARTED:
        if has_auth_timers:
            AuthTimer.objects.create(
                details=details,
                system=system,
                planet_moon=planet_moon,
                structure=structure_type_name,
                objective=objective,
                eve_time=eve_time,
                eve_corp=notif.owner.corporation,
                corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
            )
            timer_processed = True

        if has_structure_timers:
            eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                id=parsed_text["solarSystemID"]
            )
            structure_type, _ = EveType.objects.get_or_create_esi(
                id=parsed_text["structureTypeID"]
            )
            visibility = (
                Timer.Visibility.CORPORATION
                if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
                else Timer.Visibility.UNRESTRICTED
            )
            Timer.objects.create(
                eve_solar_system=eve_solar_system,
                structure_type=structure_type,
                timer_type=Timer.Type.MOONMINING,
                objective=Timer.Objective.FRIENDLY,
                date=eve_time,
                location_details=moon.name,
                eve_corporation=notif.owner.corporation,
                eve_alliance=notif.owner.corporation.alliance,
                visibility=visibility,
                structure_name=parsed_text["structureName"],
                owner_name=notif.owner.corporation.corporation_name,
                details_notes=_timer_details_notes(notif),
            )
            timer_processed = True

    elif notif.notif_type == NotificationType.MOONMINING_EXTRACTION_CANCELLED:
        timer_processed = True
        notifications_qs = Notification.objects.filter(
            notif_type=NotificationType.MOONMINING_EXTRACTION_STARTED,
            owner=notif.owner,
            is_timer_added=True,
            timestamp__lte=notif.timestamp,
        ).order_by("-timestamp")
        for notification in notifications_qs:
            parsed_text_2 = notification.parsed_text()
            my_structure_type_id = parsed_text_2["structureTypeID"]
            if my_structure_type_id == parsed_text["structureTypeID"]:
                eve_time = ldap_time_2_datetime(parsed_text_2["readyTime"])
                if has_auth_timers:
                    timer_query = AuthTimer.objects.filter(
                        system=system,
                        planet_moon=planet_moon,
                        structure=structure_type_name,
                        objective=objective,
                        eve_time=eve_time,
                    )
                    deleted_count, _ = timer_query.delete()
                    logger.info(
                        f"{notif.notification_id}: removed {deleted_count} "
                        "obsolete Auth timers related to notification"
                    )

                if has_structure_timers:
                    timer_query = Timer.objects.filter(
                        eve_solar_system=eve_solar_system,
                        structure_type=structure_type,
                        timer_type=Timer.Type.MOONMINING,
                        location_details=moon.name,
                        date=eve_time,
                        objective=Timer.Objective.FRIENDLY,
                        eve_corporation=notif.owner.corporation,
                        eve_alliance=notif.owner.corporation.alliance,
                        visibility=visibility,
                        structure_name=parsed_text["structureName"],
                        owner_name=notif.owner.corporation.corporation_name,
                    )
                    deleted_count, _ = timer_query.delete()
                    logger.info(
                        f"{notif.notification_id}: removed {deleted_count} "
                        "obsolete structure timers related to notification"
                    )

    return timer_processed


def _gen_timer_tower_reinforcements(notif: Notification, parsed_text: str) -> bool:
    """Generate timer for tower reinforcements."""
    structure = notif.structures.first()
    eve_time = dt.datetime.fromisoformat(notif.details["reinforced_until"])
    timer_processed = False
    if has_auth_timers:
        structure_type_map = {
            starbases.StarbaseSize.SMALL: "POS[S]",
            starbases.StarbaseSize.MEDIUM: "POS[M]",
            starbases.StarbaseSize.LARGE: "POS[L]",
        }
        structure_str = structure_type_map.get(
            starbases.starbase_size(structure.eve_type), "POS[M]"
        )
        AuthTimer.objects.create(
            details=gettext("Final timer"),
            system=structure.eve_solar_system.name,
            planet_moon=structure.eve_moon.name,
            structure=structure_str,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if has_structure_timers:
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=structure.eve_solar_system.id
        )
        structure_type, _ = EveType.objects.get_or_create_esi(id=structure.eve_type.id)
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=eve_solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            location_details=structure.eve_moon.name,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name=structure.name,
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _timer_details_notes(notif: Notification) -> str:
    """Return generated details notes string for Timers."""
    return (
        "Automatically created from structure notification for "
        f"{notif.owner.corporation} at {notif.timestamp.strftime(DATETIME_FORMAT)}"
    )



from django.core.management.base import BaseCommand, CommandError
from eveuniverse.models import EveEntity, EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.esi import fetch_esi_status
from app_utils.logging import LoggerAddTag

from ... import __title__
from ...models import Notification, StarbaseDetailFuel, Structure, StructureItem

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = "Preload missing eveuniverse objects in preparation for migrating to eveuniverse with release 2."

    def handle(self, *args, **options):
        if not fetch_esi_status().is_ok:
            raise CommandError("ESI currently not available. Aborting.")
        logger.info("Running command for preloading eveuniverse objects.")
        models_map = [
            (Notification, "sender", EveEntity),
            (StarbaseDetailFuel, "eve_type", EveType),
            (Structure, "eve_moon", EveMoon),
            (Structure, "eve_planet", EvePlanet),
            (Structure, "eve_solar_system", EveSolarSystem),
            (Structure, "eve_type", EveType),
            (StructureItem, "eve_type", EveType),
        ]
        for (
            StructuresModel,
            structures_field,
            EveuniverseModel,
        ) in models_map:
            ids_target = {
                value
                for value in StructuresModel.objects.values_list(
                    f"{structures_field}_id", flat=True
                )
                if value is not None
            }
            ids_current = set(EveuniverseModel.objects.values_list("id", flat=True))
            ids_diff = ids_target.difference(ids_current)
            if ids_diff:
                logger.info("%s: Missing IDs: %s", StructuresModel.__name__, ids_diff)
                self.stdout.write(
                    f"{StructuresModel.__name__}: Need to fetch {len(ids_diff)} "
                    "missing object(s) from ESI."
                )
            else:
                logger.info("%s: No missing IDs", StructuresModel.__name__)
                self.stdout.write(f"{StructuresModel.__name__}: OK")
            EveuniverseModel.objects.bulk_get_or_create_esi(ids=ids_target)
        logger.info("Preloading eveuniverse objects completed")
        self.stdout.write(self.style.SUCCESS("DONE"))



"""This custom migration ensures that all eveuniverse objects exist in the database,
which are needed in the next migration.
"""

from django.db import migrations


def ensure_eveuniverse_objects_exist(apps, schema_editor):
    models_map = [
        ("Notification", "sender", "EveEntity"),
        ("StarbaseDetailFuel", "eve_type", "EveType"),
        ("Structure", "eve_moon", "EveMoon"),
        ("Structure", "eve_planet", "EvePlanet"),
        ("Structure", "eve_solar_system", "EveSolarSystem"),
        ("Structure", "eve_type", "EveType"),
        ("StructureItem", "eve_type", "EveType"),
    ]
    for structures_model_name, structures_field, eveuniverse_model_name in models_map:
        StructuresModel = apps.get_model("structures", structures_model_name)
        needed_ids = {
            value
            for value in StructuresModel.objects.values_list(
                f"{structures_field}_id", flat=True
            )
            if value is not None
        }
        EveUniverseModel = apps.get_model("eveuniverse", eveuniverse_model_name)
        existing_ids = set(EveUniverseModel.objects.values_list("id", flat=True))
        if not needed_ids.issubset(existing_ids):
            missing_ids = needed_ids.difference(existing_ids)
            raise RuntimeError(
                "Migration to 2.x can not proceed, because you are missing "
                f"Eveuniverse objects for {StructuresModel.__name__}.{structures_field} "
                "in your database. "
                "Please run the structures_preload_eveuniverse command "
                "to load those missing objects as instructed in the update notes. "
                "Then retry to migrate. "
                f"Missing ids are: {missing_ids}"
            )


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0043_remove_services_localization"),
    ]

    operations = [
        migrations.RunPython(
            ensure_eveuniverse_objects_exist, migrations.RunPython.noop
        ),
    ]



"""Squashed migrations based on a new migration from scratch to get a clean basis."""

import multiselectfield.db.fields

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models

import structures.webhooks.core


class Migration(migrations.Migration):

    replaces = [
        ("structures", "0001_initial"),
        ("structures", "0002_auto_20191121_2335"),
        ("structures", "0003_auto_20191125_2157"),
        ("structures", "0004_auto_20191129_0308"),
        ("structures", "0005_notification_created"),
        ("structures", "0006_auto_20200120_0147"),
        ("structures", "0007_auto_20200123_1429"),
        ("structures", "0008_auto_20200209_1638"),
        ("structures", "0009_localizations"),
        ("structures", "0010_sovereignty_map"),
        ("structures", "0011_last_online"),
        ("structures", "0012_disable_pings"),
        ("structures", "0013_group_pings"),
        ("structures", "0014_addtl_perm_unanchor"),
        ("structures", "0015_refactor_notification_types"),
        ("structures", "0016_char_join_leave_notifications"),
        ("structures", "0017_war_dec_notifications"),
        ("structures", "0018_restructure_notifications_2"),
        ("structures", "0019_more_notification_types"),
        ("structures", "0020_more_war_notifications"),
        ("structures", "0021_add_corporation_permission"),
        ("structures", "0022_add_corp_app_notification_types"),
        ("structures", "0023_structure_fittings"),
        ("structures", "0024_fix_name_fields_too_small"),
        ("structures", "0025_add_poco_details"),
        ("structures", "0026_rework_update_status"),
        ("structures", "0027_create_owner_characters"),
        ("structures", "0028_migrate_owner_characters"),
        ("structures", "0029_custom_fuel_notifications"),
        ("structures", "0030_update_refueled_notifications"),
        ("structures", "0031_add_service_up"),
        ("structures", "0032_add_jump_fuel_alerts"),
        ("structures", "0033_fix_notification_type_fuel_alert"),
        ("structures", "0034_add_structure_webhooks"),
        ("structures", "0035_add_more_notifications_types"),
        ("structures", "0036_django4_update"),
        ("structures", "0037_add_billing_notification_types"),
        ("structures", "0038_add_starbase_detail"),
        ("structures", "0039_ownercharacter_error_count"),
        ("structures", "0040_add_generated_notifications"),
        ("structures", "0041_notification_sender_can_be_null"),
        ("structures", "0042_migrate_sender_category_other"),
        ("structures", "0043_remove_services_localization"),
        ("structures", "0044_check_eveuniverse_fully_loaded"),
        ("structures", "0045_migrate_all_to_django_eveuniverse"),
    ]

    initial = True

    dependencies = [
        ("eveonline", "0015_factions"),
        ("authentication", "0019_merge_20211026_0919"),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("eveuniverse", "0007_evetype_description"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app and view public pages"),
                    ("view_corporation_structures", "Can view corporation structures"),
                    ("view_alliance_structures", "Can view alliance structures"),
                    ("view_all_structures", "Can view all structures"),
                    ("add_structure_owner", "Can add new structure owner"),
                    (
                        "view_all_unanchoring_status",
                        "Can view unanchoring timers for all structures the user can see",
                    ),
                    ("view_structure_fit", "Can view structure fit"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="EveSovereigntyMap",
            fields=[
                (
                    "solar_system_id",
                    models.PositiveIntegerField(primary_key=True, serialize=False),
                ),
                (
                    "alliance_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="alliance who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "corporation_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="corporation who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "faction_id",
                    models.PositiveIntegerField(
                        blank=True,
                        db_index=True,
                        help_text="faction who holds sov for this system",
                        null=True,
                    ),
                ),
                (
                    "last_updated",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=None,
                        help_text="When this object was last updated from ESI",
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FuelAlertConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "channel_ping_type",
                    models.CharField(
                        choices=[("NO", "none"), ("HE", "here"), ("EV", "everyone")],
                        default="HE",
                        help_text="Option to ping every member of the channel. This setting can be overruled by the respective owner or webhook configuration",
                        max_length=2,
                        verbose_name="channel pings",
                    ),
                ),
                (
                    "color",
                    models.IntegerField(
                        blank=True,
                        choices=[
                            (14242639, "danger"),
                            (6013150, "info"),
                            (6076508, "success"),
                            (15773006, "warning"),
                        ],
                        default=15773006,
                        help_text="Context color of these notification on Discord",
                        null=True,
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Disabled configurations will not create any new alerts.",
                    ),
                ),
                (
                    "end",
                    models.PositiveIntegerField(
                        help_text="End of alerts in hours before fuel expires"
                    ),
                ),
                (
                    "repeat",
                    models.PositiveIntegerField(
                        help_text="Notifications will be repeated every x hours. Set to 0 for no repeats"
                    ),
                ),
                (
                    "start",
                    models.PositiveIntegerField(
                        help_text="Start of alerts in hours before fuel expires"
                    ),
                ),
            ],
            options={
                "verbose_name": "structure fuel alert config",
            },
        ),
        migrations.CreateModel(
            name="JumpFuelAlertConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "channel_ping_type",
                    models.CharField(
                        choices=[("NO", "none"), ("HE", "here"), ("EV", "everyone")],
                        default="HE",
                        help_text="Option to ping every member of the channel. This setting can be overruled by the respective owner or webhook configuration",
                        max_length=2,
                        verbose_name="channel pings",
                    ),
                ),
                (
                    "color",
                    models.IntegerField(
                        blank=True,
                        choices=[
                            (14242639, "danger"),
                            (6013150, "info"),
                            (6076508, "success"),
                            (15773006, "warning"),
                        ],
                        default=15773006,
                        help_text="Context color of these notification on Discord",
                        null=True,
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Disabled configurations will not create any new alerts.",
                    ),
                ),
                (
                    "threshold",
                    models.PositiveIntegerField(
                        help_text="Notifications will be sent once fuel level in units reaches this threshold"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        help_text="Corporation owning structures",
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="structure_owner",
                        serialize=False,
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "are_pocos_public",
                    models.BooleanField(
                        default=False,
                        help_text="whether pocos of this owner are shown on public POCO page",
                    ),
                ),
                (
                    "assets_last_update_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last successful update happened",
                        null=True,
                    ),
                ),
                (
                    "forwarding_last_update_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last successful update happened",
                        null=True,
                    ),
                ),
                (
                    "has_default_pings_enabled",
                    models.BooleanField(
                        default=True,
                        help_text="to enable or disable pinging of notifications for this owner e.g. with @everyone and @here",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="whether this owner is currently included in the sync process",
                    ),
                ),
                (
                    "is_alliance_main",
                    models.BooleanField(
                        default=False,
                        help_text="whether alliance wide notifications are forwarded for this owner (e.g. sov notifications)",
                    ),
                ),
                (
                    "is_included_in_service_status",
                    models.BooleanField(
                        default=True,
                        help_text="whether the sync status of this owner is included in the overall status of this services",
                    ),
                ),
                (
                    "is_up",
                    models.BooleanField(
                        default=None,
                        editable=False,
                        help_text="whether all services for this owner are currently up",
                        null=True,
                    ),
                ),
                (
                    "notifications_last_update_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last successful update happened",
                        null=True,
                    ),
                ),
                (
                    "structures_last_update_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last successful update happened",
                        null=True,
                    ),
                ),
                (
                    "character_ownership",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="OUTDATED. Has been replaced by OwnerCharacter",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "ping_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Groups to be pinged for each notification. ",
                        related_name="+",
                        to="auth.group",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="StarbaseDetail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("allow_alliance_members", models.BooleanField()),
                ("allow_corporation_members", models.BooleanField()),
                (
                    "anchor_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                ("attack_if_at_war", models.BooleanField()),
                ("attack_if_other_security_status_dropping", models.BooleanField()),
                (
                    "attack_security_status_threshold",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "attack_standing_threshold",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "fuel_bay_take_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "fuel_bay_view_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "last_modified_at",
                    models.DateTimeField(
                        help_text="When data was modified on the server."
                    ),
                ),
                (
                    "offline_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "online_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                (
                    "unanchor_role",
                    models.CharField(
                        choices=[
                            ("AL", "alliance member"),
                            ("CE", "config starbase equipment role"),
                            ("CO", "corporation member"),
                            ("FT", "starbase fuel technician role"),
                        ],
                        max_length=2,
                    ),
                ),
                ("use_alliance_standings", models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name="Structure",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        help_text="The Item ID of the structure",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        help_text="date this structure was received from ESI for the first time",
                    ),
                ),
                (
                    "fuel_expires_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date on which the structure will run out of fuel",
                        null=True,
                    ),
                ),
                (
                    "has_fitting",
                    models.BooleanField(
                        blank=True,
                        db_index=True,
                        default=None,
                        help_text="bool indicating if the structure has a fitting",
                        null=True,
                    ),
                ),
                (
                    "has_core",
                    models.BooleanField(
                        blank=True,
                        db_index=True,
                        default=None,
                        help_text="bool indicating if the structure has a quantum core",
                        null=True,
                    ),
                ),
                (
                    "last_online_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="date this structure had any of it's services online",
                        null=True,
                    ),
                ),
                (
                    "last_updated_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="date this structure was last updated from the EVE server",
                        null=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The full name of the structure", max_length=255
                    ),
                ),
                (
                    "next_reinforce_hour",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                        null=True,
                        validators=[django.core.validators.MaxValueValidator(23)],
                    ),
                ),
                (
                    "next_reinforce_apply",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                        null=True,
                    ),
                ),
                (
                    "reinforce_hour",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="The hour of day that determines the four hour window when the structure will randomly exit its reinforcement periods and become vulnerable to attack against its armor and/or hull. The structure will become vulnerable at a random time that is +/- 2 hours centered on the value of this property",
                        null=True,
                        validators=[django.core.validators.MaxValueValidator(23)],
                    ),
                ),
                (
                    "position_x",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="x position in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_y",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="y position in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_z",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="z position in the solar system",
                        null=True,
                    ),
                ),
                (
                    "state",
                    models.IntegerField(
                        blank=True,
                        choices=[
                            (1, "anchor vulnerable"),
                            (2, "anchoring"),
                            (3, "armor reinforce"),
                            (4, "armor vulnerable"),
                            (5, "deploy vulnerable"),
                            (6, "fitting invulnerable"),
                            (7, "hull reinforce"),
                            (8, "hull vulnerable"),
                            (9, "online deprecated"),
                            (10, "onlining vulnerable"),
                            (11, "shield vulnerable"),
                            (12, "unanchored"),
                            (21, "offline"),
                            (22, "online"),
                            (23, "onlining"),
                            (24, "reinforced"),
                            (25, "unanchoring "),
                            (0, "N/A"),
                            (13, "unknown"),
                        ],
                        default=13,
                        help_text="Current state of the structure",
                    ),
                ),
                (
                    "state_timer_end",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure entered it’s current state",
                        null=True,
                    ),
                ),
                (
                    "state_timer_start",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure will move to it’s next state",
                        null=True,
                    ),
                ),
                (
                    "unanchors_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure will unanchor",
                        null=True,
                    ),
                ),
                (
                    "eve_moon",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Moon next to this structure - if any",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to="eveuniverse.evemoon",
                    ),
                ),
                (
                    "eve_planet",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Planet next to this structure - if any",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to="eveuniverse.eveplanet",
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="type of the structure",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that owns the structure",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="structures",
                        to="structures.owner",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="StructureTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="name of the tag - must be unique",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="description for this tag",
                        null=True,
                    ),
                ),
                (
                    "style",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("default", "grey"),
                            ("primary", "dark blue"),
                            ("success", "green"),
                            ("info", "light blue"),
                            ("warning", "orange"),
                            ("danger", "red"),
                        ],
                        default="default",
                        help_text="color style of tag",
                        max_length=16,
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(
                        blank=True,
                        default=100,
                        help_text="number defining the order tags are shown. custom tags can not have an order below 100",
                        validators=[django.core.validators.MinValueValidator(100)],
                    ),
                ),
                (
                    "is_default",
                    models.BooleanField(
                        default=False,
                        help_text="if true this custom tag will automatically be added to new structures",
                    ),
                ),
                (
                    "is_user_managed",
                    models.BooleanField(
                        default=True,
                        help_text="if False this tag is created and managed by the system and can not be modified by users",
                    ),
                ),
            ],
            options={
                "ordering": ["order", "name"],
            },
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="short name to identify this webhook",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="you can add notes about this webhook here if you want",
                        null=True,
                    ),
                ),
                (
                    "webhook_type",
                    models.IntegerField(
                        choices=[(1, "Discord Webhook")],
                        default=1,
                        help_text="type of this webhook",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="whether notifications are currently sent to this webhook",
                    ),
                ),
                (
                    "notification_types",
                    multiselectfield.db.fields.MultiSelectField(
                        choices=[
                            ("StructureAnchoring", "Upwell structure anchoring"),
                            ("StructureOnline", "Upwell structure went online"),
                            (
                                "StructureServicesOffline",
                                "Upwell structure services went offline",
                            ),
                            (
                                "StructureWentHighPower",
                                "Upwell structure went high power",
                            ),
                            (
                                "StructureWentLowPower",
                                "Upwell structure went low power",
                            ),
                            ("StructureUnanchoring", "Upwell structure unanchoring"),
                            ("StructureFuelAlert", "Upwell structure fuel alert"),
                            ("StructureRefueledExtra", "Upwell structure refueled"),
                            (
                                "StructureJumpFuelAlert",
                                "Upwell structure jump fuel alert",
                            ),
                            (
                                "StructureUnderAttack",
                                "Upwell structure is under attack",
                            ),
                            ("StructureLostShields", "Upwell structure lost shields"),
                            ("StructureLostArmor", "Upwell structure lost armor"),
                            ("StructureDestroyed", "Upwell structure destroyed"),
                            (
                                "StructuresReinforcementChanged",
                                "Upwell structure reinforcement time changed",
                            ),
                            (
                                "OwnershipTransferred",
                                "Upwell structure ownership transferred",
                            ),
                            ("OrbitalAttacked", "Customs office attacked"),
                            ("OrbitalReinforced", "Customs office reinforced"),
                            ("TowerAlertMsg", "Starbase attacked"),
                            ("TowerResourceAlertMsg", "Starbase fuel alert"),
                            ("TowerRefueledExtra", "Starbase refueled (BETA)"),
                            ("TowerReinforcedExtra", "Starbase reinforced (BETA)"),
                            (
                                "MoonminingExtractionStarted",
                                "Moonmining extraction started",
                            ),
                            ("MoonminingLaserFired", "Moonmining laser fired"),
                            (
                                "MoonminingExtractionCancelled",
                                "Moonmining extraction cancelled",
                            ),
                            (
                                "MoonminingExtractionFinished",
                                "Moonmining extraction finished",
                            ),
                            (
                                "MoonminingAutomaticFracture",
                                "Moonmining automatic fracture triggered",
                            ),
                            (
                                "SovStructureReinforced",
                                "Sovereignty structure reinforced",
                            ),
                            (
                                "SovStructureDestroyed",
                                "Sovereignty structure destroyed",
                            ),
                            (
                                "EntosisCaptureStarted",
                                "Sovereignty entosis capture started",
                            ),
                            (
                                "SovCommandNodeEventStarted",
                                "Sovereignty command node event started",
                            ),
                            (
                                "SovAllClaimAquiredMsg",
                                "Sovereignty claim acknowledgment",
                            ),
                            ("SovAllClaimLostMsg", "Sovereignty lost"),
                            (
                                "AllAnchoringMsg",
                                "Structure anchoring in alliance space",
                            ),
                            ("WarDeclared", "War declared"),
                            ("AllyJoinedWarAggressorMsg", "War ally joined aggressor"),
                            ("AllyJoinedWarAllyMsg", "War ally joined ally"),
                            ("AllyJoinedWarDefenderMsg", "War ally joined defender"),
                            ("WarAdopted", "War adopted"),
                            ("WarInherited", "War inherited"),
                            ("CorpWarSurrenderMsg", "War party surrendered"),
                            ("WarRetractedByConcord", "War retracted by Concord"),
                            (
                                "CorpBecameWarEligible",
                                "War corporation became eligible",
                            ),
                            (
                                "CorpNoLongerWarEligible",
                                "War corporation no longer eligible",
                            ),
                            ("WarSurrenderOfferMsg", "War surrender offered"),
                            ("CorpAppNewMsg", "Character submitted application"),
                            (
                                "CorpAppInvitedMsg",
                                "Character invited to join corporation",
                            ),
                            ("CorpAppRejectCustomMsg", "Corp application rejected"),
                            ("CharAppWithdrawMsg", "Character withdrew application"),
                            ("CharAppAcceptMsg", "Character joins corporation"),
                            ("CharLeftCorpMsg", "Character leaves corporation"),
                            ("BillOutOfMoneyMsg", "Bill out of money"),
                            (
                                "InfrastructureHubBillAboutToExpire",
                                "I-HUB bill about to expire",
                            ),
                            (
                                "IHubDestroyedByBillFailure",
                                "I_HUB destroyed by bill failure",
                            ),
                        ],
                        default=[
                            "StructureAnchoring",
                            "StructureDestroyed",
                            "StructureFuelAlert",
                            "StructureLostArmor",
                            "StructureLostShields",
                            "StructureOnline",
                            "StructureServicesOffline",
                            "StructureUnderAttack",
                            "StructureWentHighPower",
                            "StructureWentLowPower",
                            "OrbitalAttacked",
                            "OrbitalReinforced",
                            "TowerAlertMsg",
                            "TowerResourceAlertMsg",
                            "SovStructureReinforced",
                            "SovStructureDestroyed",
                        ],
                        help_text="select which type of notifications should be forwarded to this webhook",
                        max_length=1123,
                    ),
                ),
                (
                    "language_code",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("en", "English"),
                            ("de", "German"),
                            ("es", "Spanish"),
                            ("zh-hans", "Chinese Simplified"),
                            ("ru", "Russian"),
                            ("ko", "Korean"),
                        ],
                        default=None,
                        help_text="language of notifications send to this webhook",
                        max_length=8,
                        null=True,
                        verbose_name="language",
                    ),
                ),
                (
                    "is_default",
                    models.BooleanField(
                        default=False,
                        help_text="whether owners have this webhook automatically pre-set when created",
                    ),
                ),
                (
                    "has_default_pings_enabled",
                    models.BooleanField(
                        default=True,
                        help_text="to enable or disable pinging of notifications for this webhook e.g. with @everyone and @here",
                    ),
                ),
                (
                    "ping_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Groups to be pinged for each notification - ",
                        related_name="+",
                        to="auth.group",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(structures.webhooks.core.DiscordWebhookMixin, models.Model),
        ),
        migrations.CreateModel(
            name="StructureItem",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        help_text="The Eve item ID", primary_key=True, serialize=False
                    ),
                ),
                ("is_singleton", models.BooleanField()),
                ("last_updated_at", models.DateTimeField(auto_now=True)),
                ("location_flag", models.CharField(max_length=255)),
                ("quantity", models.IntegerField()),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="eve type of the item",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "structure",
                    models.ForeignKey(
                        help_text="Structure this item is located in",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="items",
                        to="structures.structure",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="structure",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="List of tags for this structure. ",
                related_name="structures",
                to="structures.structuretag",
            ),
        ),
        migrations.AddField(
            model_name="structure",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Webhooks for sending notifications to. If any webhook is enabled, these will be used instead of the webhooks defined for the respective owner. If no webhook is enabled the owner's setting will be used. ",
                related_name="structures",
                to="structures.webhook",
            ),
        ),
        migrations.CreateModel(
            name="StarbaseDetailFuel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.IntegerField()),
                (
                    "detail",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fuels",
                        to="structures.starbasedetail",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="starbasedetail",
            name="structure",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="starbase_detail",
                to="structures.structure",
            ),
        ),
        migrations.CreateModel(
            name="PocoDetails",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("alliance_tax_rate", models.FloatField(default=None, null=True)),
                ("allow_access_with_standings", models.BooleanField()),
                ("allow_alliance_access", models.BooleanField()),
                ("bad_standing_tax_rate", models.FloatField(default=None, null=True)),
                ("corporation_tax_rate", models.FloatField(default=None, null=True)),
                (
                    "excellent_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                ("good_standing_tax_rate", models.FloatField(default=None, null=True)),
                (
                    "neutral_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                ("reinforce_exit_end", models.PositiveIntegerField()),
                ("reinforce_exit_start", models.PositiveIntegerField()),
                (
                    "standing_level",
                    models.IntegerField(
                        choices=[
                            (-99, "none"),
                            (-10, "terrible"),
                            (-5, "bad"),
                            (0, "neutral"),
                            (5, "good"),
                            (10, "excellent"),
                        ],
                        default=-99,
                    ),
                ),
                (
                    "terrible_standing_tax_rate",
                    models.FloatField(default=None, null=True),
                ),
                (
                    "structure",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="poco_details",
                        to="structures.structure",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="OwnerCharacter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "structures_last_used_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        editable=False,
                        help_text="when this character was last used for syncing structures",
                        null=True,
                    ),
                ),
                (
                    "notifications_last_used_at",
                    models.DateTimeField(
                        db_index=True,
                        default=None,
                        editable=False,
                        help_text="when this character was last used for syncing notifications",
                        null=True,
                    ),
                ),
                (
                    "error_count",
                    models.PositiveIntegerField(
                        default=0,
                        editable=False,
                        help_text="Count of ESI errors which happened with this character.",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "character_ownership",
                    models.ForeignKey(
                        help_text="character used for syncing",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="characters",
                        to="structures.owner",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="owner",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Notifications are sent to these webhooks. ",
                related_name="owners",
                to="structures.webhook",
            ),
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_sent",
                    models.BooleanField(
                        default=False,
                        help_text="True when this notification has been forwarded to Discord",
                    ),
                ),
                (
                    "is_timer_added",
                    models.BooleanField(
                        default=False,
                        help_text="True when a timer has been added for this notification",
                        null=True,
                    ),
                ),
                (
                    "notif_type",
                    models.CharField(
                        db_index=True,
                        default="",
                        help_text="type of this notification",
                        max_length=100,
                        verbose_name="type",
                    ),
                ),
                ("notification_id", models.PositiveBigIntegerField(verbose_name="id")),
                (
                    "created",
                    models.DateTimeField(
                        default=None,
                        help_text="Date when this notification was first received from ESI",
                        null=True,
                    ),
                ),
                (
                    "is_read",
                    models.BooleanField(
                        default=None,
                        help_text="True when this notification has read in the eve client",
                        null=True,
                    ),
                ),
                (
                    "last_updated",
                    models.DateTimeField(
                        help_text="Date when this notification has last been updated from ESI"
                    ),
                ),
                (
                    "text",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="Notification details in YAML",
                        null=True,
                    ),
                ),
                ("timestamp", models.DateTimeField(db_index=True)),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that owns this notification",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.owner",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "structures",
                    models.ManyToManyField(
                        help_text="Structures this notification is about (if any)",
                        to="structures.structure",
                    ),
                ),
            ],
            options={
                "verbose_name": "eve notification",
            },
        ),
        migrations.CreateModel(
            name="JumpFuelAlert",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "config",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jump_fuel_alerts",
                        to="structures.jumpfuelalertconfig",
                    ),
                ),
                (
                    "structure",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jump_fuel_alerts",
                        to="structures.structure",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="GeneratedNotification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_sent",
                    models.BooleanField(
                        default=False,
                        help_text="True when this notification has been forwarded to Discord",
                    ),
                ),
                (
                    "is_timer_added",
                    models.BooleanField(
                        default=False,
                        help_text="True when a timer has been added for this notification",
                        null=True,
                    ),
                ),
                (
                    "notif_type",
                    models.CharField(
                        db_index=True,
                        default="",
                        help_text="type of this notification",
                        max_length=100,
                        verbose_name="type",
                    ),
                ),
                ("details", models.JSONField(default=dict)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that owns this notification",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.owner",
                    ),
                ),
                (
                    "structures",
                    models.ManyToManyField(
                        help_text="Structures this notification is about (if any)",
                        to="structures.structure",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="FuelAlert",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "hours",
                    models.PositiveIntegerField(
                        db_index=True,
                        help_text="number of hours before fuel expiration this alert was sent",
                    ),
                ),
                (
                    "config",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="structure_fuel_alerts",
                        to="structures.fuelalertconfig",
                    ),
                ),
                (
                    "structure",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="structure_fuel_alerts",
                        to="structures.structure",
                    ),
                ),
            ],
            options={
                "verbose_name": "structure fuel alert",
            },
        ),
        migrations.CreateModel(
            name="StructureService",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(help_text="Name of the service", max_length=100),
                ),
                (
                    "state",
                    models.IntegerField(
                        choices=[(1, "offline"), (2, "online")],
                        help_text="Current state of this service",
                    ),
                ),
                (
                    "structure",
                    models.ForeignKey(
                        help_text="Structure this service is installed to",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="services",
                        to="structures.structure",
                    ),
                ),
            ],
            options={
                "unique_together": {("structure", "name")},
            },
        ),
        migrations.AddConstraint(
            model_name="ownercharacter",
            constraint=models.UniqueConstraint(
                fields=("owner", "character_ownership"), name="functional_pk_ownertoken"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="notification",
            unique_together={("notification_id", "owner")},
        ),
        migrations.AddConstraint(
            model_name="fuelalert",
            constraint=models.UniqueConstraint(
                fields=("structure", "config", "hours"), name="functional_pk_fuelalert"
            ),
        ),
    ]



# Generated by Django 4.0.3 on 2022-04-09 18:52

import multiselectfield.db.fields

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0036_django4_update"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webhook",
            name="notification_types",
            field=multiselectfield.db.fields.MultiSelectField(
                choices=[
                    ("StructureAnchoring", "Upwell structure anchoring"),
                    ("StructureOnline", "Upwell structure went online"),
                    (
                        "StructureServicesOffline",
                        "Upwell structure services went offline",
                    ),
                    ("StructureWentHighPower", "Upwell structure went high power"),
                    ("StructureWentLowPower", "Upwell structure went low power"),
                    ("StructureUnanchoring", "Upwell structure unanchoring"),
                    ("StructureFuelAlert", "Upwell structure fuel alert"),
                    ("StructureRefueledExtra", "Upwell structure refueled"),
                    ("StructureJumpFuelAlert", "Upwell structure jump fuel alert"),
                    ("StructureUnderAttack", "Upwell structure is under attack"),
                    ("StructureLostShields", "Upwell structure lost shields"),
                    ("StructureLostArmor", "Upwell structure lost armor"),
                    ("StructureDestroyed", "Upwell structure destroyed"),
                    (
                        "StructuresReinforcementChanged",
                        "Upwell structure reinforcement time changed",
                    ),
                    ("OwnershipTransferred", "Upwell structure ownership transferred"),
                    ("OrbitalAttacked", "Customs office attacked"),
                    ("OrbitalReinforced", "Customs office reinforced"),
                    ("TowerAlertMsg", "Starbase attacked"),
                    ("TowerResourceAlertMsg", "Starbase fuel alert"),
                    ("TowerRefueledExtra", "Starbase refueled"),
                    ("MoonminingExtractionStarted", "Moonmining extraction started"),
                    ("MoonminingLaserFired", "Moonmining laser fired"),
                    (
                        "MoonminingExtractionCancelled",
                        "Moonmining extraction cancelled",
                    ),
                    ("MoonminingExtractionFinished", "Moonmining extraction finished"),
                    (
                        "MoonminingAutomaticFracture",
                        "Moonmining automatic fracture triggered",
                    ),
                    ("SovStructureReinforced", "Sovereignty structure reinforced"),
                    ("SovStructureDestroyed", "Sovereignty structure destroyed"),
                    ("EntosisCaptureStarted", "Sovereignty entosis capture started"),
                    (
                        "SovCommandNodeEventStarted",
                        "Sovereignty command node event started",
                    ),
                    ("SovAllClaimAquiredMsg", "Sovereignty claim acknowledgment"),
                    ("SovAllClaimLostMsg", "Sovereignty lost"),
                    ("AllAnchoringMsg", "Structure anchoring in alliance space"),
                    ("WarDeclared", "War declared"),
                    ("AllyJoinedWarAggressorMsg", "War ally joined aggressor"),
                    ("AllyJoinedWarAllyMsg", "War ally joined ally"),
                    ("AllyJoinedWarDefenderMsg", "War ally joined defender"),
                    ("WarAdopted", "War adopted"),
                    ("WarInherited", "War inherited"),
                    ("CorpWarSurrenderMsg", "War party surrendered"),
                    ("WarRetractedByConcord", "War retracted by Concord"),
                    ("CorpBecameWarEligible", "War corporation became eligable"),
                    ("CorpNoLongerWarEligible", "War corporation no longer eligable"),
                    ("WarSurrenderOfferMsg", "War surrender offered"),
                    ("CorpAppNewMsg", "Character submitted application"),
                    ("CorpAppInvitedMsg", "Character invited to join corporation"),
                    ("CorpAppRejectCustomMsg", "Corp application rejected"),
                    ("CharAppWithdrawMsg", "Character withdrew application"),
                    ("CharAppAcceptMsg", "Character joins corporation"),
                    ("CharLeftCorpMsg", "Character leaves corporation"),
                    ("BillOutOfMoneyMsg", "Bill out of money"),
                    (
                        "InfrastructureHubBillAboutToExpire",
                        "I-HUB bill about to expire",
                    ),
                    ("IHubDestroyedByBillFailure", "I_HUB destroyed by bill failure"),
                ],
                default=[
                    "StructureAnchoring",
                    "StructureDestroyed",
                    "StructureFuelAlert",
                    "StructureLostArmor",
                    "StructureLostShields",
                    "StructureOnline",
                    "StructureServicesOffline",
                    "StructureUnderAttack",
                    "StructureWentHighPower",
                    "StructureWentLowPower",
                    "OrbitalAttacked",
                    "OrbitalReinforced",
                    "TowerAlertMsg",
                    "TowerResourceAlertMsg",
                    "SovStructureReinforced",
                    "SovStructureDestroyed",
                ],
                help_text="select which type of notifications should be forwarded to this webhook",
                max_length=1102,
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-11-21 03:27

import multiselectfield.db.fields

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import structures.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0010_alliance_ticker"),
        ("authentication", "0016_ownershiprecord"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app and view"),
                    ("view_alliance_structures", "Can view alliance structures"),
                    ("view_all_structures", "Can view all structures"),
                    ("add_structure_owner", "Can add new structure owner"),
                ),
                "default_permissions": (),
                "managed": False,
            },
        ),
        migrations.CreateModel(
            name="EveConstellation",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online region ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="EveEntity",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                (
                    "category",
                    models.IntegerField(
                        choices=[
                            (1, "character"),
                            (2, "corporation"),
                            (3, "alliance"),
                            (4, "faction"),
                            (5, "other"),
                        ]
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveGroup",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online group ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="EveRegion",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online region ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="EveSolarSystem",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online solar system ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("security_status", models.FloatField()),
                (
                    "eve_constellation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveConstellation",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveType",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online type ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "eve_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveGroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        help_text="Corporation owning structures",
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "structures_last_sync",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last sync happened",
                        null=True,
                    ),
                ),
                (
                    "structures_last_error",
                    models.IntegerField(
                        choices=[
                            (0, "No error"),
                            (1, "Invalid token"),
                            (2, "Expired token"),
                            (3, "Insufficient permissions"),
                            (4, "No character set for fetching alliance contacts"),
                            (5, "ESI API is currently unavailable"),
                            (6, "Operaton mode does not match with current setting"),
                            (99, "Unknown error"),
                        ],
                        default=0,
                        help_text="error that occurred at the last sync atttempt (if any)",
                    ),
                ),
                (
                    "notifications_last_sync",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="when the last sync happened",
                        null=True,
                    ),
                ),
                (
                    "notifications_last_error",
                    models.IntegerField(
                        choices=[
                            (0, "No error"),
                            (1, "Invalid token"),
                            (2, "Expired token"),
                            (3, "Insufficient permissions"),
                            (4, "No character set for fetching alliance contacts"),
                            (5, "ESI API is currently unavailable"),
                            (6, "Operaton mode does not match with current setting"),
                            (99, "Unknown error"),
                        ],
                        default=0,
                        help_text="error that occurred at the last sync atttempt (if any)",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="character used for syncing structures",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to="authentication.CharacterOwnership",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Structure",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        help_text="The Item ID of the structure",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The full name of the structure", max_length=255
                    ),
                ),
                (
                    "position_x",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="x position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_y",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="y position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_z",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="z position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "fuel_expires",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date on which the structure will run out of fuel",
                        null=True,
                    ),
                ),
                (
                    "next_reinforce_hour",
                    models.IntegerField(
                        blank=True,
                        default=None,
                        help_text="The requested change to reinforce_hour that will take effect at the time shown by next_reinforce_apply",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(23),
                        ],
                    ),
                ),
                (
                    "next_reinforce_weekday",
                    models.IntegerField(
                        blank=True,
                        default=None,
                        help_text="The date and time when the structure’s newly requested reinforcement times (e.g. next_reinforce_hour and next_reinforce_day) will take effect",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(6),
                        ],
                    ),
                ),
                (
                    "next_reinforce_apply",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="The requested change to reinforce_weekday that will take effect at the time shown by next_reinforce_apply",
                        null=True,
                    ),
                ),
                (
                    "reinforce_hour",
                    models.IntegerField(
                        help_text="The hour of day that determines the four hour window when the structure will randomly exit its reinforcement periods and become vulnerable to attack against its armor and/or hull. The structure will become vulnerable at a random time that is +/- 2 hours centered on the value of this property",
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(23),
                        ],
                    ),
                ),
                (
                    "reinforce_weekday",
                    models.IntegerField(
                        blank=True,
                        default=None,
                        help_text="The day of the week when the structure exits its final reinforcement period and becomes vulnerable to attack against its hull. Monday is 0 and Sunday is 6",
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(6),
                        ],
                    ),
                ),
                (
                    "state",
                    models.IntegerField(
                        choices=[
                            (0, "N/A"),
                            (1, "anchor_vulnerable"),
                            (2, "anchoring"),
                            (3, "armor_reinforce"),
                            (4, "armor_vulnerable"),
                            (5, "deploy_vulnerable"),
                            (6, "fitting_invulnerable"),
                            (7, "hull_reinforce"),
                            (8, "hull_vulnerable"),
                            (9, "online_deprecated"),
                            (10, "onlining_vulnerable"),
                            (11, "shield_vulnerable"),
                            (12, "unanchored"),
                            (13, "unknown"),
                        ],
                        help_text="Current state of the structure",
                    ),
                ),
                (
                    "state_timer_start",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure will move to it’s next state",
                        null=True,
                    ),
                ),
                (
                    "state_timer_end",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure entered it’s current state",
                        null=True,
                    ),
                ),
                (
                    "unanchors_at",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="Date at which the structure will unanchor",
                        null=True,
                    ),
                ),
                (
                    "last_updated",
                    models.DateTimeField(
                        blank=True,
                        default=None,
                        help_text="date this structure was last updated from the EVE server",
                        null=True,
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveSolarSystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="type of the structure",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveType",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that owns the structure",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.Owner",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="short name to identify this webhook",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "webhook_type",
                    models.IntegerField(
                        choices=[(1, "Discord Webhook")],
                        default=1,
                        help_text="type of this webhook",
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="you can add notes about this webhook here if you want",
                        null=True,
                    ),
                ),
                (
                    "notification_types",
                    multiselectfield.db.fields.MultiSelectField(
                        choices=[
                            (401, "MoonminingAutomaticFracture"),
                            (402, "MoonminingExtractionCancelled"),
                            (403, "MoonminingExtractionFinished"),
                            (404, "MoonminingExtractionStarted"),
                            (405, "MoonminingLaserFired"),
                            (513, "OwnershipTransferred"),
                            (501, "StructureAnchoring"),
                            (502, "StructureDestroyed"),
                            (503, "StructureFuelAlert"),
                            (504, "StructureLostArmor"),
                            (505, "StructureLostShields"),
                            (506, "StructureOnline"),
                            (507, "StructureServicesOffline"),
                            (508, "StructureUnanchoring"),
                            (509, "StructureUnderAttack"),
                            (510, "StructureWentHighPower"),
                            (511, "StructureWentLowPower"),
                        ],
                        default=structures.models.get_default_notification_types,
                        help_text="only notifications which selected types are sent to this webhook",
                        max_length=67,
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="whether notifications are currently sent to this webhook",
                    ),
                ),
                (
                    "is_default",
                    models.BooleanField(
                        default=False,
                        help_text="whether newly added owners have this automatically webhook preset",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="owner",
            name="webhooks",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="notifications are sent to these webhooks. ",
                to="structures.Webhook",
            ),
        ),
        migrations.CreateModel(
            name="EveMoon",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online item ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "position_x",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="x position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_y",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="y position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_z",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="z position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveSolarSystem",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="eveconstellation",
            name="eve_region",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="structures.EveRegion"
            ),
        ),
        migrations.CreateModel(
            name="StructureService",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(help_text="Name of the service", max_length=64),
                ),
                (
                    "state",
                    models.IntegerField(
                        choices=[(1, "offline"), (2, "online")],
                        help_text="Current state of this service",
                    ),
                ),
                (
                    "structure",
                    models.ForeignKey(
                        help_text="Structure this service is installed to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.Structure",
                    ),
                ),
            ],
            options={
                "unique_together": {("structure", "name")},
            },
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "notification_id",
                    models.BigIntegerField(
                        validators=[django.core.validators.MinValueValidator(0)]
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                (
                    "notification_type",
                    models.IntegerField(
                        choices=[
                            (401, "MoonminingAutomaticFracture"),
                            (402, "MoonminingExtractionCancelled"),
                            (403, "MoonminingExtractionFinished"),
                            (404, "MoonminingExtractionStarted"),
                            (405, "MoonminingLaserFired"),
                            (513, "OwnershipTransferred"),
                            (501, "StructureAnchoring"),
                            (502, "StructureDestroyed"),
                            (503, "StructureFuelAlert"),
                            (504, "StructureLostArmor"),
                            (505, "StructureLostShields"),
                            (506, "StructureOnline"),
                            (507, "StructureServicesOffline"),
                            (508, "StructureUnanchoring"),
                            (509, "StructureUnderAttack"),
                            (510, "StructureWentHighPower"),
                            (511, "StructureWentLowPower"),
                        ]
                    ),
                ),
                ("text", models.TextField(blank=True, default=None, null=True)),
                ("is_read", models.BooleanField(blank=True, default=None, null=True)),
                ("is_sent", models.BooleanField(blank=True, default=False)),
                ("last_updated", models.DateTimeField()),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Corporation that received this notification",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.Owner",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveEntity",
                    ),
                ),
            ],
            options={
                "unique_together": {("notification_id", "owner")},
            },
        ),
    ]



# Generated by Django 2.2.8 on 2020-01-23 14:29

import multiselectfield.db.fields

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import structures.models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0006_auto_20200120_0147"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="notification_type",
            field=models.IntegerField(
                choices=[
                    (401, "MoonminingAutomaticFracture"),
                    (402, "MoonminingExtractionCancelled"),
                    (403, "MoonminingExtractionFinished"),
                    (404, "MoonminingExtractionStarted"),
                    (405, "MoonminingLaserFired"),
                    (513, "OwnershipTransferred"),
                    (501, "StructureAnchoring"),
                    (502, "StructureDestroyed"),
                    (503, "StructureFuelAlert"),
                    (504, "StructureLostArmor"),
                    (505, "StructureLostShields"),
                    (506, "StructureOnline"),
                    (507, "StructureServicesOffline"),
                    (508, "StructureUnanchoring"),
                    (509, "StructureUnderAttack"),
                    (510, "StructureWentHighPower"),
                    (511, "StructureWentLowPower"),
                    (601, "OrbitalAttacked"),
                    (602, "OrbitalReinforced"),
                ]
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="notification_types",
            field=multiselectfield.db.fields.MultiSelectField(
                choices=[
                    (401, "MoonminingAutomaticFracture"),
                    (402, "MoonminingExtractionCancelled"),
                    (403, "MoonminingExtractionFinished"),
                    (404, "MoonminingExtractionStarted"),
                    (405, "MoonminingLaserFired"),
                    (513, "OwnershipTransferred"),
                    (501, "StructureAnchoring"),
                    (502, "StructureDestroyed"),
                    (503, "StructureFuelAlert"),
                    (504, "StructureLostArmor"),
                    (505, "StructureLostShields"),
                    (506, "StructureOnline"),
                    (507, "StructureServicesOffline"),
                    (508, "StructureUnanchoring"),
                    (509, "StructureUnderAttack"),
                    (510, "StructureWentHighPower"),
                    (511, "StructureWentLowPower"),
                    (601, "OrbitalAttacked"),
                    (602, "OrbitalReinforced"),
                ],
                default=structures.models.get_default_notification_types,
                help_text="only notifications which selected types are sent to this webhook",
                max_length=75,
            ),
        ),
        migrations.CreateModel(
            name="EvePlanet",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        help_text="Eve Online item ID",
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "position_x",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="x position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_y",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="y position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "position_z",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="z position of the structure in the solar system",
                        null=True,
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveSolarSystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="structures.EveType",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.0.7 on 2022-09-16 12:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("structures", "0001_initial_new"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="sender",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_moon",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Moon next to this structure - if any",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.evemoon",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_planet",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Planet next to this structure - if any",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.eveplanet",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_solar_system",
            field=models.ForeignKey(
                help_text="Solar System the structure is located",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evesolarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="structure",
            name="eve_type",
            field=models.ForeignKey(
                help_text="Type of the structure",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.evetype",
            ),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-02 18:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("structures", "0035_add_more_notifications_types"),
    ]

    operations = [
        migrations.AlterField(
            model_name="owner",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification. ",
                related_name="+",
                to="auth.group",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Groups to be pinged for each notification - ",
                related_name="+",
                to="auth.group",
            ),
        ),
    ]



"""Eve sovereignty related core logic."""

from typing import Optional

from ..constants import EveTypeId

_MAP_CAMPAIGN_EVENT_2_TYPE_ID = {
    1: EveTypeId.TCU,
    2: EveTypeId.IHUB,
}
_MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME = {
    EveTypeId.CUSTOMS_OFFICE: "POCO",
    EveTypeId.TCU: "TCU",
    EveTypeId.IHUB: "I-HUB",
}


def type_id_from_event_type(event_type: int) -> Optional[int]:
    return _MAP_CAMPAIGN_EVENT_2_TYPE_ID.get(event_type)


def structure_type_name_from_event_type(event_type: int) -> str:
    return _MAP_TYPE_ID_2_TIMER_STRUCTURE_NAME.get(
        type_id_from_event_type(event_type), "Other"
    )
